#include <iostream>
#include <vector>
#include <fstream>
#include <sstream>
#include <string>
#include <functional>
#include <algorithm>
#include <cctype>
#include <list>
#include <ctime>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <boost/crc.hpp>

using namespace std;

typedef unsigned int uint32;
typedef unsigned short uint16;
typedef unsigned char uint8;
typedef uint8 byte;

#if defined (__GNUC__) && (defined (CPU_x86) || defined (CPU_x86_64))
// Optimized routines for x86{_64}
static inline const uint16 swap16 (const uint16 x)
{
	uint16 ret;
	__asm ("xchgb %%al,%%ah" : "=a" (ret) : "a" (x));
	return ret;
}
static inline const uint32 swap32 (const uint32 x)
{
	uint32 ret;
	__asm ("bswapl %%eax" : "=a" (ret) : "a" (x));
	return ret;
}
#else
static inline const uint16 swap16 (const uint16 x)
{ return (x >> 8) | (x << 8); }
static inline const uint32 swap32 (const uint32 x)
{ return (x >> 24) | ((x >> 8) & 0xff00) | ((x << 8) & 0xff0000) | (x << 24); }
#endif

inline uint32 AlignTo4(uint32 alignMe)
{
	while (alignMe % 4 !=0)
	{
		alignMe++;
	}
	return alignMe;
}

#pragma pack (push, 8)

vector<byte> makeUBFI(uint32 load,uint32 ep,uint8 os,uint8 arch,uint8 type,string name,list< vector<byte> > &subImages)
{
	typedef struct 
	{
		uint32        ih_magic;				/* Image Header Magic Number    */
		uint32        ih_hcrc;				/* Image Header CRC Checksum    */
		uint32        ih_time;				/* Image Creation Timestamp     */
		uint32        ih_size;				/* Image Data Size              */
		uint32        ih_load;				/* Data  Load  Address          */
		uint32        ih_ep;				/* Entry Point Address          */
		uint32        ih_dcrc;				/* Image Data CRC Checksum      */
		uint8         ih_os;				/* Operating System             */
		uint8         ih_arch;				/* CPU architecture             */
		uint8         ih_type;				/* Image Type                   */
		uint8         ih_comp;				/* Compression Type             */
		uint8         ih_name[32];			/* Image Name           */
	} image_header;

	image_header theHeader;
	memset(&theHeader,0,sizeof(theHeader));
	theHeader.ih_magic = swap32(0x27051956);
	theHeader.ih_time = swap32(uint32(time(NULL)));
	theHeader.ih_load = swap32(load);
	theHeader.ih_ep = swap32(ep);
	theHeader.ih_os = os;
	theHeader.ih_arch = arch;
	theHeader.ih_type = type;
	theHeader.ih_comp = 0;
	if (name.length() > 0)
		strncpy((char*)theHeader.ih_name,name.c_str(),sizeof(theHeader.ih_name)-1);

	vector<byte> resultVector;

	vector<uint32> subImageSizes;
	for (list< vector<byte> >::const_iterator i=subImages.begin();i!=subImages.end();++i)
	{
		subImageSizes.push_back( swap32( AlignTo4(i->size()) ) );
	}
	subImageSizes.push_back(0);

	resultVector.resize(sizeof(theHeader)+subImageSizes.size()*sizeof(uint32),0);
	memcpy(&resultVector[sizeof(theHeader)],&subImageSizes[0],subImageSizes.size()*sizeof(uint32));

	for (list< vector<byte> >::const_iterator i=subImages.begin();i!=subImages.end();++i)
	{
		std::copy(i->begin(),i->end(),std::back_inserter(resultVector));
		while(resultVector.size()%4 != 0)
		{
			resultVector.push_back(0);
		}
	}

	uint32 imageSize = resultVector.size()-sizeof(theHeader);
	theHeader.ih_size = swap32(imageSize);
	boost::crc_32_type dataCrc;
	dataCrc.process_bytes(&resultVector[sizeof(theHeader)],imageSize);
	theHeader.ih_dcrc = swap32(dataCrc.checksum());

	boost::crc_32_type headerCrc;
	headerCrc.process_bytes((const byte*)&theHeader,sizeof(theHeader));
	theHeader.ih_hcrc = swap32(headerCrc.checksum());

	memcpy(&resultVector[0],&theHeader,sizeof(theHeader));

	return resultVector;
}

void strReplace(string& str, const string oldStr, const string newStr)
{
	size_t pos = 0;
	while((pos = str.find(oldStr, pos)) != string::npos)
	{
		str.replace(pos, oldStr.length(), newStr);
		pos += newStr.length();
	}
}

void showHelp()
{
	cout	<<"mkimage -s <scripttxt> -ck <kernelbin> (-ms <squashfsbin>|-us <squashfsdir>) -o <outputbin>" << endl
			<<"-ck stands for compressed kernel" << endl
			<<"-ms stands for premade squashfs, it will take squashfs from binary file, -us for unmade filesystem, squashfs bin will be made from this folder" << endl
			<<"quickly hacked together by rajkosto" << endl;
}
int main(int argc, char *argv[])
{
	if (argc != 9)
	{
		showHelp();
		return 0;
	}
	stringstream cmdlist;
	for (int i=1;i<argc;i++)
	{
		cmdlist << argv[i] << " ";
	}
	string scriptFileName;
	bool scriptDefined = false;
	string kernelFileName;
	bool kernelDefined = false;
	typedef enum { SQUASH_MADE, SQUASH_UNMADE } squashType;
	squashType squashOptionType = SQUASH_MADE;
	string squashFileName;
	bool squashDefined = false;
	string outputFileName;
	bool outputDefined = false;
	while (cmdlist.eof() == false)
	{
		string currCmd;
		cmdlist >> currCmd;
		transform(currCmd.begin(), currCmd.end(), currCmd.begin(), ::tolower);
		if (currCmd == "-s")
		{
			if (scriptDefined == true)
			{
				cout << "Script already defined" << endl;
				return 0;
			}
			cmdlist >> scriptFileName;
			scriptDefined = true;
		}
		else if (currCmd == "-ck")
		{
			if (kernelDefined == true)
			{
				cout << "Kernel already defined" << endl;
				return 0;
			}

			cmdlist >> kernelFileName;
			kernelDefined = true;
		}
		else if (currCmd == "-ms" | currCmd == "-us")
		{
			if (squashDefined == true)
			{
				cout << "SquashFS already defined" << endl;
				return 0;
			}

			if (currCmd == "-ms")
				squashOptionType = SQUASH_MADE;
			else if (currCmd == "-us")
				squashOptionType = SQUASH_UNMADE;

			cmdlist >> squashFileName;
			squashDefined = true;
		}
		else if (currCmd == "-o")
		{
			if (outputDefined == true)
			{
				cout << "Output already defined" << endl;
				return 0;
			}

			cmdlist >> outputFileName;
			outputDefined = true;
		}
	}
	if (scriptDefined == false || kernelDefined == false || squashDefined == false || outputDefined == false ||
		scriptFileName.length() < 1 || kernelFileName.length() < 1 || squashFileName.length() < 1 || outputFileName.length() < 1)
	{
		cout << "Missing or invalid parameters" << endl;
		showHelp();
	}

	cout << "Script File: " << scriptFileName << endl;
	cout << "Kernel File: " << kernelFileName << endl;
	cout << "SquashFS File/Dir: " << squashFileName << " ";
	if (squashOptionType == SQUASH_MADE)
	{
		cout << "(From already created file)" << endl;
	}
	else if (squashOptionType == SQUASH_UNMADE)
	{
		cout << "(Will be created from dir)" << endl;
	}
	cout << "Output File: " << outputFileName << endl;

	ifstream currFile;
	currFile.open(scriptFileName.c_str(),ios::in|ios::binary);
	if (currFile.is_open() == false)
	{
		cout << "Error opening script file" << endl;
		return 0;
	}
	currFile.seekg(0,ios::end);
	uint32 fileSize = currFile.tellg();
	currFile.seekg(0,ios::beg);
	vector <byte> scriptVect(fileSize);
	currFile.read((char*)&scriptVect[0],scriptVect.size());
	currFile.close();

	//replace CRLF in script file with just LF for linux compat
	{
		string scriptContents((char*)&scriptVect[0],scriptVect.size());
		strReplace(scriptContents,"\r\n","\n");
		scriptVect.resize(scriptContents.size());
		memcpy(&scriptVect[0],scriptContents.data(),scriptContents.size());
	}

	list < vector<byte> > tempList;
	tempList.push_back(scriptVect);
	vector<byte> scriptUBFI = makeUBFI(0,0,5,7,6,"Boot Script File",tempList);

	currFile.open(kernelFileName.c_str(),ios::in|ios::binary);
	if (currFile.is_open() == false)
	{
		cout << "Error opening kernel file" << endl;
		return 0;
	}
	currFile.seekg(0,ios::end);
	fileSize = currFile.tellg();
	currFile.seekg(0,ios::beg);
	vector<byte> kernelVect(fileSize);
	currFile.read((char*)&kernelVect[0],kernelVect.size());
	currFile.close();

	vector<byte> squashVect;
	if (squashOptionType == SQUASH_UNMADE)
	{
		string systemcmd = string("mksquashfs ") + squashFileName + string(" tempsquash.bin -be");
		if (system(systemcmd.c_str()) != 0)
		{
			cout << "Error running mksquashfs command" << endl;
			return 0;
		}

		squashFileName = "tempsquash.bin";

		currFile.open(squashFileName.c_str(),ios::in|ios::binary);
		if (currFile.is_open() == false)
		{
			cout << "Error opening temporary squashfs file" << endl;
			return 0;
		}
		currFile.seekg(0,ios::end);
		uint32 newSize = currFile.tellg();
		if (newSize == 0)
		{
			currFile.close();
			cout << "Squashfs is empty" << endl;
			return 0;
		}

		currFile.seekg(0,ios::beg);
		squashVect.resize(newSize);
		currFile.read((char*)&squashVect[0],newSize);
		currFile.close();
		unlink(squashFileName.c_str());

		cout << "Generated SquashFS is " << squashVect.size() << " bytes" << endl;
	}
	else
	{
		currFile.open(squashFileName.c_str(),ios::in|ios::binary);
		if (currFile.is_open() == false)
		{
			cout << "Error opening SquashFS file" << endl;
			return 0;
		}
		currFile.seekg(0,ios::end);
		fileSize = currFile.tellg();
		currFile.seekg(0,ios::beg);
		squashVect.resize(fileSize);
		currFile.read((char*)&squashVect[0],squashVect.size());
		currFile.close();
	}

	tempList.clear();
	tempList.push_back(kernelVect);
	tempList.push_back(squashVect);

	vector<byte> firmwareUBFI = makeUBFI(0x80A00000,0x80A00000,5,2,4,"Multi Image File",tempList);

	ofstream outputFile;
	outputFile.open(outputFileName.c_str(),ios::out|ios::binary|ios::trunc);
	if (outputFile.is_open() == false)
	{
		cout << "Error opening output file" << endl;
		return 0;
	}
	outputFile.write((char*)&scriptUBFI[0],scriptUBFI.size());
	outputFile.write((char*)&firmwareUBFI[0],firmwareUBFI.size());
	fileSize = outputFile.tellp();
	outputFile.close();
	cout << "Written " << fileSize << " bytes" << endl;
	
	return 0;
}

#pragma pack (pop)
